/**
 * 
 * given a rip structure, calculate its energy
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include "energy.h"
#include "pfunc.h"
#include "calc_energy.h"

paramT *P;
int *int_r, *int_s;             /* the integer array representation of rip structure */
int *strand_r, *strand_s;       /* judge an arc is on the same strand or not */
int len_r, len_s;                   /* length of $struc_r and $struc_s */
int num_align;                      /* number of sequences */

int exist_interaction(int i, int flag);
double calc_TS_energy(int i, int j, int r, int s, char **seqs_r, char **seqs_s);
double calc_DT_energy(int i, int j, int r, int s, char **seqs_r, char **seqs_s);
double calc_RT_energy(int i, int j, int r, int s, char **seqs_r, char **seqs_s);


void initialize_array(void)
{
    int i, *ptr;
    
    if (num_align==0 || len_r==0 || len_s==0) {
        fprintf(stderr, "%s:%d: error: initialization failure. No sequences.\n", __FILE__, __LINE__);
        exit(EXIT_FAILURE);
    }

    int_r = (int *) space(sizeof(int)*(len_r+1));
    int_s = (int *) space(sizeof(int)*(len_s+1));
    strand_r = (int *) space(sizeof(int)*(len_r+1));
    strand_s = (int *) space(sizeof(int)*(len_s+1));
    
    memset(strand_r, 0, sizeof(int)*(len_r+1));
    memset(strand_s, 0, sizeof(int)*(len_s+1));
    int_r[0] = len_r;
    int_s[0] = len_s;

    aSR = (int **) space(sizeof(int *) * (num_align+1));
    ptr = (int *) space(sizeof(int) * (num_align+1) * (len_r+1));
    for (i=0; i<(num_align+1); i++) {
        aSR[i] = ptr;
        ptr += len_r+1;
    }
    
    aSS = (int **) space(sizeof(int *) * (num_align+1));
    ptr = (int *) space(sizeof(int) * (num_align+1) * (len_s+1)) ;
    for (i=0; i<(num_align+1); i++) {
        aSS[i] = ptr;
        ptr += len_s+1;
    }

	P = scale_parameters();
    
    if (num_align == 0) {
        fprintf(stderr, "File %s, line %d: error to calculate kTn!\n", __FILE__, __LINE__);
        exit(EXIT_FAILURE);
    }

    return;
}

void free_array(void)
{
    free(int_r); free(int_s);
    free(strand_r); free(strand_s);

    free(aSR[0]); free(aSR);
	free(aSS[0]); free(aSS);

    int_r = int_s = NULL;
    strand_r = strand_s = NULL;
}
    
/**
 *
 * for a block, find its positions of secondary structure
 *
 */
void find_ss_pos(int i, int j, int r, int s, int *p, int *q, int flag)
{
    int p1, p2;
    
    if (flag == LEFT) {         /* find position from left */
        p1 = i;
        p2 = r;

        //R
        while (p1 <= j) {
            if (int_r[p1] == 0) {
                p1++;
                continue;
            }

            if (strand_r[p1]==SS || (strand_r[p1]==SR && exist_interaction(p1,SR))) {
                *p = p1;
                break;
            }

            p1 = int_r[p1];
            p1++;
        }
        if (p1 > j)
            *p = j+1;

        //S
        while (p2 <= s) {
            if (int_s[len_s-p2+1] == 0) {
                p2++;
                continue;
            }

            if (strand_s[len_s-p2+1]==SR || (strand_s[len_s-p2+1]==SS && exist_interaction(p2,SS))) {
                *q = p2;
                break;
            }

            p2 = len_s-int_s[len_s-p2+1]+1;
            p2++;
        }
        if (p2 > s)             /* structure on S is secondary */
            *p = s+1;
    }
    else {                    /* find position from right */
        p1 = j;
        p2 = s;

        //R
        while (p1 >= i) {
            if (int_r[p1] == 0) {
                p1--;
                continue;
            }

            if (strand_r[p1]==SS || (strand_r[p1]==SR && exist_interaction(p1,SR))) {
                *p = p1;
                break;
            }

            p1 = int_r[p1];
            p1--;
        }
        if (p1 < i)             /* structure on R is secondary */
            *p = i-1;

        //S
        while (p2 >= r) {
            if (int_s[len_s-p2+1] == 0) {
                p2--;
                continue;
            }

            if (strand_s[len_s-p2+1]==SR || (strand_s[len_s-p2+1]==SS && exist_interaction(p2,SS))) {
                *q = p2;
                break;
            }

            p2 = len_s-int_s[len_s-p2+1]+1;
            p2--;
        }
        if (p2 < r)             /* structure on S is secondary */
            *p = r-1;
    }

    return;
}

double stack_energy(int i, int *pair_table, int **aS, char **string)
{
    /* calculate energy of substructure enclosed by (i,j) */
    int ee;
    double energy = 0.0;
    int unpaired = 0;
    int j, p, q, k;
    int *type, *type_2;

    type = (int *) space(sizeof(int)*(num_align+1));
    type_2 = (int *) space(sizeof(int)*(num_align+1));
    
    j = pair_table[i];
    for (k=1; k<=num_align; k++) {
        type[k] = BP_pair[aS[k][i]][aS[k][j]];
        if (type[k] == 0)
            type[k] = 7;
    }

    p = i;
    q = j;
    while (p < q) { /* process all stacks and interior loops */
        ee = 0;
        
        while (pair_table[++p] == 0)
            unpaired++;
        while (pair_table[--q] == 0)
            unpaired++;
        if ((pair_table[q]!=p) || (p>q))
            break;
        
        for (k=1; k<=num_align; k++) {
            type_2[k] = BP_pair[aS[k][q]][aS[k][p]];
            if (type_2[k] == 0)
                type_2[k] = 7;
            ee += calc_interior_energy(p-i-1, j-q-1, type[k], type_2[k],
                                       aS[k][i+1], aS[k][j-1], aS[k][p-1], aS[k][q+1]);
        }
        
        energy += ((double) ee)/num_align;
        i = p;
        j = q;
        for (k=1; k<=num_align; k++)
            type[k] = rtype[type_2[k]];
    } /* end while */

    /* p,q don't pair must have found hairpin or multiloop */
    if (p > q) {                       /* hairpin */
        ee = 0;
        
        for (k=1; k<=num_align; k++)
            ee += calc_hairpin_energy(string[k]+i-1, j-i-1, type[k], aS[k][i+1], aS[k][j-1]);
        energy += ((double) ee)/num_align;
        
        return energy;
    }

    /* (i,j) is exterior pair of multiloop */
    while (p < j) {
        /* add up the contributions of the substructures of the ML */
        energy += stack_energy(p, pair_table, aS, string);

        ee = 0;
        for (k=1; k<=num_align; k++)
            ee += P->MLintern[BP_pair[aS[k][p]][aS[k][pair_table[p]]]];
        energy += ((double) ee)/num_align;
        
        p = pair_table[p];
        /* search for next base pair in multiloop */
        while (pair_table[++p] == 0)
            unpaired++;
    }

    ee = 0;
    for (k=1; k<=num_align; k++)
        ee += P->MLintern[BP_pair[aS[k][i]][aS[k][j]]];
    energy += ((double) ee)/num_align;
        
    energy += P->MLclosing+P->MLbase*unpaired;
    //energy += ee;

    free(type);
    free(type_2);
    
    return energy;
}

double calc_sec_energy(int p, int q, int **aS, int *pair_table, char **seqs, int flag)
{
    int i, s, t;
    double energy = 0.0;

    if (p > q) {
        fprintf(stderr, "%s,%d: error: invalid block of secondary structure.\n", __FILE__, __LINE__);
        exit(EXIT_FAILURE);
    }    

    if ((q-p) == 0)            /* no secondary structure */
        return 0;

    if (flag == SR) {
        s = p;
        t = q;
    } else {
        s = len_s-q+1;
        t = len_s-p+1;
    }
    
    for (i=s; i<=t; i++) {
        if (pair_table[i] == 0)
            continue;
        energy += stack_energy(i, pair_table, aS, seqs);
        i = pair_table[i];
    }
    
    return energy;
}

/**
 *
 * calculate the bonus energy, i.e. kissing loop bonus and multiloop bonus.
 * i is paired to some base on the same strand. If the current loop is kissing loop,
 * then return its bonus energy; otherwise, return the bonus energy of multiloop.
 * 
 */

/* double calc_bonus_energy(int i, int flag) */
/* { */
/*     int j, p, k; */
/*     int from, to; */
/*     int unpaired=0, n_loop=0; */
/*     double ml_energy=0.0, ks_energy=0.0;        */
/*     int loop_type = 0;    /\* 0 if multiloop and 1 kissing loop *\/ */
/*     int ee; */

/*     if (flag == SR) { */
/*         j = int_r[i]; */
        
/*         if (strand_r[i] == SS) { */
/*             fprintf(stderr, "%s,%d: error: can not start with an interaction in judging kissing loop.\n", */
/*                     __FILE__, __LINE__); */
/*             exit(EXIT_FAILURE); */
/*         } */

/*         from = (i<j)?i:j; */
/*         to = (i<j)?j:i; */
/*         for (p=from+1; p<to; p++) { */
/*             if (int_r[p] == 0) { */
/*                 unpaired++; */
/*                 continue; */
/*             } */
            
/*             /\* interaction or arc on R *\/ */
/*             if (strand_r[p] == SS) */
/*                 loop_type = 1; */
/*             else { */
/*                 ee = 0; */
                
/*                 for (k=1; k<=num_align; k++) */
/*                     ee += P->MLintern[BP_pair[aSR[k][p]][aSR[k][int_r[p]]]]; */
/*                 ml_energy += ((double) ee)/num_align; */
                
/*                 p = int_r[p]; */
/*                 n_loop++; */
/*             } */
/*         }    */
/*     } */
/*     else { */
/*         j = int_s[len_s-i+1]; */

/*         if (strand_s[len_s-i+1] == SR) { */
/*             fprintf(stderr, "%s,%d: error: can not start with an interaction in judging kissing loop.\n", */
/*                     __FILE__, __LINE__); */
/*             exit(EXIT_FAILURE); */
/*         } */
        
/*         from = ((len_s-i+1)<j) ? (len_s-i+1) : j; */
/*         to = ((len_s-i+1)<j) ? j : (len_s-i+1); */
/*         for (p=from+1; p<to; p++) { */
/*             if (int_s[p] == 0) { */
/*                 unpaired++; */
/*                 continue; */
/*             } */

/*             /\* interaction or arc on S *\/ */
/*             if (strand_s[p] == SR) */
/*                 loop_type = 1; */
/*             else { */
/*                 ee = 0; */
                
/*                 for (k=1; k<=num_align; k++) */
/*                     ee += P->MLintern[BP_pair[aSS[k][p]][aSS[k][int_s[p]]]]; */
/*                 ml_energy += ((double) ee)/num_align; */
                
/*                 p = int_s[p]; */
/*                 n_loop++; */
/*             } */
/*         } */
/*     } */

/*     if (loop_type == 1) {   /\* kissing loop *\/ */
/*         ks_energy = BETA1+BETA2*n_loop+BETA3*unpaired; */
/*         return ks_energy; */
/*     } */
/*     else {                      /\* multiloop *\/ */
/*         ee = 0; */

/*         if (flag == SR) { */
/*             for (k=1; k<=num_align; k++) */
/*                 ee += P->MLintern[BP_pair[aSR[k][from]][aSR[k][to]]]; */
/*         } */
/*         else { */
/*             for (k=1; k<=num_align; k++) */
/*                 ee += P->MLintern[BP_pair[aSS[k][from]][aSS[k][to]]]; */
/*         } */

/*         ml_energy += P->MLclosing+((double) ee)/num_align+P->MLbase*unpaired; */
/*         return ml_energy; */
/*     } */
/* } */


int exist_interaction(int i, int flag)
{
	int j, k;
    int from, to;
	
    if (flag == SR) {
        j = int_r[i];

        if (strand_r[i] == SS) {
            fprintf(stderr, "%s,%d: error: can not start with an interaction in judging the existence of interaction.\n",
                    __FILE__, __LINE__);
            exit(EXIT_FAILURE);
        }
        
        from = (i<j)?i:j;
        to = (i<j)?j:i;
        for (k=from+1; k<to; k++) {
            if (strand_r[k] == SS)
                return 1;       /* find an interaction */
        }
    } else {
        j = int_s[len_s-i+1];

        if (strand_s[len_s-i+1] == SR) {
            fprintf(stderr, "%s,%d: error: can not start with an interaction in judging the existence of interaction.\n",
                    __FILE__, __LINE__);
            exit(EXIT_FAILURE);
        }
        
        from = ((len_s-i+1)<j)?(len_s-i+1):j;
        to = ((len_s-i+1)<j)?j:(len_s-i+1);
        for (k=from+1; k<to; k++) {
            if (strand_s[k] == SR)
                return 1;       /* find an interaction */
        }
    }

    return 0;
}

int is_interaction(int r, int s)
{
    if (strand_r[r]==SR || strand_s[len_s-s+1]==SS)
        return 0;
    if (int_r[r]!=len_s-s+1 || int_s[len_s-s+1]!=r)
        return 0;

    return 1;
}

double calc_DT_energy(int i, int j, int r, int s, char **seqs_r, char **seqs_s)
{
    int k, t, io, ro;
    int p, q, i1, r1;
    int ee, dt_energy = 0;
    int *type, *type_2, type_s;
    int nr_inter, ns_inter;

    if (i>j || r>s)
        return 0.0;

    type = (int *) space(sizeof(int)*(num_align+1));
    type_2 = (int *) space(sizeof(int)*(num_align+1));
    
    if (is_interaction(i,r)) {  /* hybrid+rt or hybrid */
        for (k=1; k<=num_align; k++) {
            type[k] = BP_pair[aSR[k][i]][aSS[k][len_s-r+1]];
            if (type[k] == 0)
                type[k] = 7;
        }

        /* search hybrid */
        io = i; ro = r;
        p = i; q = r;
        while (p<j && q<s) {
            ee = 0;
            
            while (int_r[++p] == 0);
            while (int_s[len_s-(++q)+1] == 0);
            if (!is_interaction(p,q)) /* if there is single interaction, the energy contribution is 0 */
                break;
            
            for (k=1; k<=num_align; k++) {
                type_2[k] = BP_pair[aSS[k][len_s-q+1]][aSR[k][p]];
                if (type_2[k] == 0)
                    type_2[k] = 7;
                ee += calc_interior_energy(p-i-1, q-r-1, type[k], type_2[k],
                                           aSR[k][i+1], aSS[k][len_s-r], aSR[k][p-1], aSS[k][len_s-q+2]);
            }
            dt_energy += SIGMA*ee/num_align;

            i = p;
            r = q;
            for (k=1; k<=num_align; k++)
                type[k] = rtype[type_2[k]];
        }

        #ifndef  DEBUG
        if (p<j && q<s)
            fprintf(stdout, "DT[%d,%d; %d,%d] = hybrid[%d,%d; %d,%d]+RT[%d,%d; %d,%d]\n",
                    io, j, ro, s, io, i, ro, r, p, j, q, s);
        else
            fprintf(stdout, "DT[%d,%d; %d,%d] = hybrid[%d,%d; %d,%d]\n",
                    io, j, ro, s, io, i, ro, r);
        #endif
        
        if (p<j && q<s)      /* there exists dt */
            dt_energy += calc_RT_energy(i+1,j,r+1,s,seqs_r,seqs_s);
    } else {                    /* ts + rt */
        /* search tight structure */
        nr_inter = ns_inter = 0;

        if (strand_r[i]==SR && strand_s[len_s-r+1]==SS) { /* Te, Tu or Td*/
            p = int_r[i];
            q = len_s-int_s[len_s-r+1]+1;
        } else if (strand_r[i] == SS) {                /* Tu*/
            p = i-1;
            q = len_s-int_s[len_s-r+1]+1;
        } else if (strand_s[len_s-r+1] == SR) { /* Td */
            p = int_r[i];
            q = r-1;
        }
        
        for (t=i+1; t<p; t++) {
            if (strand_r[t] == SS)
                nr_inter++;
        }
        for (t=(len_s-q+2); t<(len_s-r+1); t++) {
            if (strand_s[t] == SR)
                ns_inter++;
        }
        if (nr_inter == ns_inter) { /* Te */
            i1 = p;
            r1 = q;
        } else if (nr_inter > ns_inter) { /* Tu */
            i1 = p;
            while (ns_inter < nr_inter) {
                if (strand_s[len_s-(++q)+1] == SR)
                    ns_inter++;
            }
            
            /* find the rightmost index on S*/
            t = q;
            while (int_s[len_s-(++q)+1] == 0);
            if (strand_s[len_s-q+1] == SR)
                r1 = t;
            else if (int_s[len_s-q+1] < (len_s-q+1))
                r1 = t;
            else if (int_s[len_s-q+1] > (len_s-q+1)) {   /* kissing loop */
                type_s = 0;                              /* hybrid afterwards if type=1, and secondary structure if type=2*/
                
                while ((++q) <= s) {
                    if (int_s[len_s-q+1] == 0)
                        continue;
                    
                    if (strand_s[len_s-q+1] == SR) {        /* there exists interaction */
                        t = q;
                        
                        while ((++q)<=s && (int_s[len_s-q+1]==0 || strand_s[len_s-q+1]==SR));
                        if (q>s || int_s[len_s-q+1]<(len_s-q+1)) { /* hybrid */
						    q = s;
                            type_s = 1;
                            break;
                        } else { /* zigzag */
                            fprintf(stderr, "%s,%d: error: the block is not a dt.\n", __FILE__, __LINE__);
                            exit(EXIT_FAILURE);
                        }
                            
                    }

                    if (int_s[len_s-q+1] < (len_s-q+1)) { /* secondary structure */
                        type_s = 2;
                        t = q;
                        break;
                    }
                }

                if (q > s)
                    r1 = s;
                else if (type_s==1 || type_s==2) {
                    while (int_s[len_s-(--t)+1] == 0);
                    r1 = t;
                } else {        /* maybe type=0 or other error*/
                    fprintf(stderr, "%s,%d: error: an unknown error occurred.\n", __FILE__, __LINE__);
                    exit(EXIT_FAILURE);
                }
            }
        } else if (nr_inter < ns_inter) {                /* Td */
            r1 = q;
            while (nr_inter < ns_inter) {
                if (strand_r[++p] == SS)
                    nr_inter++;
            }
            
            /* find the rightmost index on R */
            t = p;
            while (int_r[++p] == 0);
            
            if (strand_r[p] == SS)
                i1 = t;
            else if (int_r[p] > p)
                i1 = t;
            else if (int_r[p] < p) {   /* kissing loop */
                type_s = 0;               /* hybrid afterwards if type=1, and secondary structure if type=2*/
                
                while ((++p) <= j) {
                    if (int_r[p] == 0)
                        continue;
                    
                    if (strand_r[p] == SS) {        /* there exists interaction */
                        t = p;
                        
                        while ((++p)<=j && (int_r[p]==0 || strand_r[p]==SS));
                        if (p>j || int_r[p]>p) { /* hybrid */
						    p = j;
                            type_s = 1;
                            break;
                        } else { /* zigzag */
                            fprintf(stderr, "%s,%d: error: the block is not a dt.\n", __FILE__, __LINE__);
                            exit(EXIT_FAILURE);
                        }
                    }

                    if (int_r[p] > p) { /* secondary structure */
                        type_s = 2;
                        t = p;
                        break;
                    }
                }

                if (p > j)
                    i1 = j;
                else if (type_s==1 || type_s==2) {
                    while (int_r[--t] == 0);
                    i1 = t;
                } else {        /* maybe type=0 or other error*/
                    fprintf(stderr, "%s,%d: error: an unknown error occurred.\n", __FILE__, __LINE__);
                    exit(EXIT_FAILURE);
                }
            }
        }

        #ifndef  DEBUG
        if (i1==j && r1==s) {
            if (int_r[i]==i1 && int_s[len_s-r1+1]==(len_s-r+1)) /* Te */
                fprintf(stdout, "DT[%d,%d; %d,%d] = Te[%d,%d; %d,%d]\n",
                        i, j, r, s, i, i1, r, r1);
            else if (int_r[i] == i1) /* Tu */
                fprintf(stdout, "DT[%d,%d; %d,%d] = Tu[%d,%d; %d,%d]\n",
                        i, j, r, s, i, i1, r, r1);
            else if (int_s[len_s-r1+1] == (len_s-r+1)) /* Td */
                fprintf(stdout, "DT[%d,%d; %d,%d] = Td[%d,%d; %d,%d]\n",
                        i, j, r, s, i, i1, r, r1);
        }
        else if (i1<j && r1<s) {
            if (int_r[i]==i1 && int_s[len_s-r1+1]==(len_s-r+1)) /* Te */
                fprintf(stdout, "DT[%d,%d; %d,%d] = Te[%d,%d; %d,%d]+RT[%d,%d; %d,%d]\n",
                        i, j, r, s, i, i1, r, r1, i1+1, j, r1+1, s);
            else if (int_r[i] == i1) /* Tu */
                fprintf(stdout, "DT[%d,%d; %d,%d] = Tu[%d,%d; %d,%d]+RT[%d,%d; %d,%d]\n",
                        i, j, r, s, i, i1, r, r1, i1+1, j, r1+1, s);
            else if (int_s[len_s-r1+1] == (len_s-r+1)) /* Td */
                fprintf(stdout, "DT[%d,%d; %d,%d] = Td[%d,%d; %d,%d]+RT[%d,%d; %d,%d]\n",
                        i, j, r, s, i, i1, r, r1, i1+1, j, r1+1, s);
        }
        #endif
        
        if (i1==j && r1==s)
            dt_energy += calc_TS_energy(i,i1,r,r1,seqs_r,seqs_s);
        else if (i1<j && r1<s)
            dt_energy += calc_TS_energy(i,i1,r,r1,seqs_r,seqs_s)+calc_RT_energy(i1+1,j,r1+1,s,seqs_r,seqs_s);
        else {
            fprintf(stderr, "%s,%d: error: this block is invalid.\n", __FILE__, __LINE__);
            exit(EXIT_FAILURE);
        }
    }

    free(type);
    free(type_2);

    return dt_energy;
}

double calc_TS_energy(int i, int j, int r, int s, char **seqs_r, char **seqs_s)
{
    int k, p, q;
    int i1, j1, r1, s1;
    int ee;
    double energy=0.0, ml_energy=0.0;
    int unpaired=0, n_loop=0;
    int *type, *type_2, block_type, b_type;
    int once=1, loop_type=0;

    if (i>j || r>s)
        return 0.0;
   
	if (strand_r[i]==SR && strand_r[j]==SR && int_r[i]==j && int_r[j]==i) { /* (i,j) paired */
	    if ((strand_s[len_s-s+1]==SS) && (strand_s[len_s-r+1]==SR) /* Te */
		    && (int_s[len_s-r+1]==(len_s-s+1)) && (int_s[len_s-s+1]==(len_s-r+1))) {
		    b_type = Te;
	    }
	    else
		    b_type = Tu;
	}
	else if ((strand_s[len_s-s+1]==SS) && (strand_s[len_s-r+1]==SS) /* Td */
			   && (int_s[len_s-r+1]==(len_s-s+1)) && (int_s[len_s-s+1]==(len_s-r+1))) {
	    b_type = Td;
	}
	else {
	    fprintf(stderr, "%s,%d: error: the current block [%d,%d; %d,%d] is not a tight structure.\n",
				  __FILE__, __LINE__, i, j, r, s);
	    exit(EXIT_FAILURE);
	}

    type = (int *) space(sizeof(int)*(num_align+1));
    type_2 = (int *) space(sizeof(int)*(num_align+1));
    
	if (b_type==Tu || b_type==Te) {        /* Tu or Te */
        int io, jo;
        
        io = i; jo = j;
        p = i; q = j;
        r1 = r; s1 = s;
        for (k=1; k<=num_align; k++) {
            type[k] = BP_pair[aSR[k][i]][aSR[k][j]];
            if (type[k] == 0)
                type[k] = 7;
        }
        
        while (p < q) {
            ee = 0;
            
            while (int_r[++p] == 0);
            while (int_r[--q] == 0);
            if ((int_r[p]!=q) || (p>q) || (strand_r[p]==SS || strand_r[q]==SS))
                break;
            
            for (k=1; k<=num_align; k++) {
                type_2[k] = BP_pair[aSR[k][q]][aSR[k][p]];
                if (type_2[k] == 0)
                    type_2[k] = 7;
                ee += calc_interior_energy(p-i-1, j-q-1, type[k], type_2[k],
                                           aSR[k][i+1], aSR[k][j-1], aSR[k][p-1], aSR[k][q+1]);
            }
            energy += ((double) ee)/num_align;
            i = p;
            j = q;
            for (k=1; k<=num_align; k++)
                type[k] = rtype[type_2[k]];
        }
        
        if (p > q) {            /* not a tight structure */
            fprintf(stderr, "%s, %d: error: the current block is not a tight structure.\n", __FILE__, __LINE__);
            exit(EXIT_FAILURE);
        }
        else {                /* multiloop or kissing loop */
            for (p=i+1; p<j; p++) {
                if (int_r[p] == 0) {
                    unpaired++;
                    continue;
                }
            
                /* interaction or arc on R */
                if (strand_r[p] == SS) {
                    j1 = p;
                    if (once) {
                        loop_type = 1;
                        i1 = p;
                        once = 0;
                    }
                }
                else {          /* arc beginning with p and ending with int_r[p] */
                    ee = 0;

                    if (exist_interaction(p,SR)) { /* whether exist interactions below the arc (p,int_r[p]) */
                        j1 = int_r[p];
                        if (once) {
                            i1 = p;
                            once = 0;
                        }
                    }
                    else        /* stack energy over (p,int_r[p]) */
                        energy += stack_energy(p, int_r, aSR, seqs_r);

                    /* multiloop energy bonus */
                    for (k=1; k<=num_align; k++)
                        ee += P->MLintern[BP_pair[aSR[k][p]][aSR[k][int_r[p]]]];
                    ml_energy += ((double) ee)/num_align;
                
                    p = int_r[p];
                    n_loop++;
                }
            }

            if (loop_type == 1) /* kissing loop */
                energy += BETA1+BETA2*n_loop+BETA3*unpaired;
            else {              /* multiloop */
                ee = 0;
                
                for (k=1; k<=num_align; k++)
                    ee += P->MLintern[BP_pair[aSR[k][i]][aSR[k][j]]];
                energy += P->MLclosing+((double) ee)/num_align+P->MLbase*unpaired+ml_energy;
            }
            
            if (strand_r[i1]==SS || strand_r[j1]==SS) { /* there is at least one interaction */
                if (int_s[len_s-s+1] == (len_s-r+1))
                    block_type = Td;
                else
                    block_type = DT;
            }
            else if (int_r[i1] == j1) {
                if (int_s[len_s-s+1] == (len_s-r+1))
                    block_type = Te;
                else
                    block_type = Tu;
            }
            else {
                if (int_s[len_s-s+1] == (len_s-r+1))
                    block_type = Td;
                else
                    block_type = DT;
            }
        }

        #ifndef  DEBUG
        if (int_s[len_s-s+1] == (len_s-r+1)) { /* Te block */
            if (block_type == Te)
                fprintf(stdout, "Te[%d,%d; %d,%d] = Te[%d,%d; %d,%d]+gap+stack\n",
                        io, jo, r, s, i1, j1, r, s);
            else if (block_type == Td)
                fprintf(stdout, "Te[%d,%d; %d,%d] = Td[%d,%d; %d,%d]+gap+stack\n",
                        io, jo, r, s, i1, j1, r, s);
        }
        else {                  /* Tu block */
            if (block_type == Tu)
                fprintf(stdout, "Tu[%d,%d; %d,%d] = Tu[%d,%d; %d,%d]+gap+stack\n",
                        io, jo, r, s, i1, j1, r, s);
            else if (block_type == DT)
                fprintf(stdout, "Tu[%d,%d; %d,%d] = DT[%d,%d; %d,%d]+gap+stack\n",
                        io, jo, r, s, i1, j1, r, s);
        }
        #endif
	}
    else {                    /* Td */
        int i53, j53, p53, q53;           /* index from 5' to 3' */
        int ro, so;

        ro = r; so = s;
        i53 = len_s-s+1;
        j53 = len_s-r+1;
        if (int_s[i53] != j53) {
            fprintf(stderr, "%s,%d: error: the current block is not Td.\n", __FILE__, __LINE__);
            exit(EXIT_FAILURE);
        }
        
        p53 = i53; q53 = j53;
        i1 = i; j1 = j;
        for (k=1; k<=num_align; k++) {
            type[k] = BP_pair[aSS[k][i53]][aSS[k][j53]];
            if (type[k] == 0)
                type[k] = 7;
        }

        while (p53 < q53) {
            ee = 0;
            
            while (int_s[++p53] == 0);
            while (int_s[--q53] == 0);
            if ((int_s[p53]!=q53) || (p53>q53) || (strand_s[p53]==SR || strand_s[q53]==SR))
                break;
            
            for (k=1; k<=num_align; k++) {
                type_2[k] = BP_pair[aSS[k][q53]][aSS[k][p53]];
                if (type_2[k] == 0)
                    type_2[k] = 7;
                ee += calc_interior_energy(p53-i53-1, j53-q53-1, type[k], type_2[k],
                                           aSS[k][i53+1], aSS[k][j53-1], aSS[k][p53-1], aSS[k][q53+1]);
            }
            energy += ((double) ee)/num_align;
            i53 = p53;
            j53 = q53;
            for (k=1; k<=num_align; k++)
                type[k] = rtype[type_2[k]];
        }

        if (p53 > q53) {            /* not a tight structure */
            fprintf(stderr, "%s, %d: error: the current block is not a tight structure.\n", __FILE__, __LINE__);
            exit(EXIT_FAILURE);
        }
        else {
            for (p=i53+1; p<j53; p++) {
                if (int_s[p] == 0) {
                    unpaired++;
                    continue;
                }
                
                /* interaction or arc on R */
                if (strand_s[p] == SR) {
                    r1 = len_s-p+1;
                    if (once) {
                        loop_type = 1;
                        s1 = len_s-p+1;
                        once = 0;
                    }
                }
                else {          /* arc beginning with p and ending with int_r[p] */
                    ee = 0;
                
                    if (exist_interaction(len_s-p+1,SS)) { /* whether exist interactions below the arc (p,int_r[p]) */
                        r1 = len_s-int_s[p]+1;
                        if (once) {
                            s1 = len_s-p+1;
                            once = 0;
                        }
                    }
                    else        /* stack energy over (p,int_r[p]) */
                        energy += stack_energy(p, int_s, aSS, seqs_s);

                    /* multiloop energy bonus */
                    for (k=1; k<=num_align; k++)
                        ee += P->MLintern[BP_pair[aSS[k][p]][aSS[k][int_s[p]]]];
                    ml_energy += ((double) ee)/num_align;
                
                    p = int_s[p];
                    n_loop++;
                }
            }

            if (loop_type == 1) /* kissing loop */
                energy += BETA1+BETA2*n_loop+BETA3*unpaired;
            else {              /* multiloop */
                ee = 0;
                
                for (k=1; k<=num_align; k++)
                    ee += P->MLintern[BP_pair[aSS[k][i53]][aSS[k][j53]]];
                energy += P->MLclosing+((double) ee)/num_align+P->MLbase*unpaired+ml_energy;
            }
            
            if (strand_s[len_s-r1+1]==SR || strand_s[len_s-s1+1]==SR) { /* there is at least one interaction */
                block_type = DT;
            } else if (int_s[len_s-s1+1] == (len_s-r1+1)) /* Td */
                block_type = Td;
            else
                block_type = DT;
        }

        #ifndef  DEBUG
        if (block_type == Td)
            fprintf(stdout, "Td[%d,%d; %d,%d] = Td[%d,%d; %d,%d]+gap+stack\n",
                    i, j, ro, so, i, j, r1, s1);
        else if (block_type == DT)
            fprintf(stdout, "Td[%d,%d; %d,%d] = DT[%d,%d; %d,%d]+gap+stack\n",
                    i, j, ro, so, i, j, r1, s1);
        #endif
    }

    if (block_type==Tu || block_type==Te || block_type==Td)
	    energy += calc_TS_energy(i1,j1,r1,s1,seqs_r,seqs_s);
    else if (block_type == DT)
        energy += calc_DT_energy(i1,j1,r1,s1,seqs_r,seqs_s);
    energy += ml_energy;

    free(type);
    free(type_2);

    return energy;
}

double calc_RT_energy(int i, int j, int r, int s, char **seqs_r, char **seqs_s)
{
	int p, q;
	double energy = 0.0;

	find_ss_pos(i,j,r,s,&p,&q,LEFT);
    
    if (i < p)
        energy += calc_sec_energy(i,p-1,aSR,int_r,seqs_r,SR);
    if (r < q)
        energy += calc_sec_energy(r,q-1,aSS,int_s,seqs_s,SS);
    
	if ((strand_r[p]==SR && int_r[p]==j) || (strand_s[len_s-q+1]==SS && int_s[len_s-q+1]==(len_s-s+1))) {  /* tight structure */
        #ifndef  DEBUG
        if (int_r[p]==j && int_s[len_s-q+1]==(len_s-s+1)) /* Te */
            fprintf(stdout, "RT[%d,%d; %d,%d] = ss[%d,%d; %d,%d]+Te[%d,%d; %d,%d]\n",
                    i, j, r, s, i, (i<p)?(p-1):-1, r, (r<q)?(q-1):-1, p, j, q, s);
        else if (int_r[p] != j) /* Td */
            fprintf(stdout, "RT[%d,%d; %d,%d] = ss[%d,%d; %d,%d]+Td[%d,%d; %d,%d]\n",
                    i, j, r, s, i, (i<p)?(p-1):-1, r, (r<q)?(q-1):-1, p, j, q, s);
        else                    /* Tu */
            fprintf(stdout, "RT[%d,%d; %d,%d] = ss[%d,%d; %d,%d]+Tu[%d,%d; %d,%d]\n",
                    i, j, r, s, i, (i<p)?(p-1):-1, r, (r<q)?(q-1):-1, p, j, q, s);
        #endif
    
		energy += calc_TS_energy(p,j,q,s,seqs_r,seqs_s);
    } else  {       /* double tight structure */
        #ifndef  DEBUG
        fprintf(stdout, "RT[%d,%d; %d,%d] = ss[%d,%d; %d,%d]+DT[%d,%d; %d,%d]\n",
                i, j, r, s, i, (i<p)?(p-1):-1, r, (r<q)?(q-1):-1, p, j, q, s);
        #endif
        
		energy += calc_DT_energy(p,j,q,s,seqs_r,seqs_s);
    }
        
	
	return energy;
}

double calc_rip_energy(char **seqs_r, char **seqs_s)
{
	int j, s;
	double energy = 0.0;
    
    find_ss_pos(1,len_r,1,len_s,&j,&s,RIGHT);

    #ifndef  DEBUG
    fprintf(stdout, "initial structure = RT[%d,%d; %d,%d]+ss[%d,%d; %d,%d]\n", 1, j, 1, s,
            (j<len_r)?(j+1):-1, len_r, (s<len_s)?(s+1):-1, len_s);
    #endif
    
    if (j <= (len_r-1))
        energy += calc_sec_energy(j+1, len_r, aSR, int_r, seqs_r, SR);
    if (s <= (len_s-1))
        energy += calc_sec_energy(s+1, len_s, aSS, int_s, seqs_s, SS);

    if (j>=1 && s>=1)
        energy += calc_RT_energy(1,j,1,s,seqs_r,seqs_s);

    return energy;
}
