#include "stk.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static pre_list *u4;
static pre_list *u6;

rna_seq *create_rna_node(char *name, char *seq)
{
    rna_seq *rna_node = (rna_seq *) malloc(sizeof(rna_seq));
    strcpy(rna_node->name, name);
    strcpy(rna_node->seq, seq);
    rna_node->next = NULL;

    return rna_node;
}

pre_node *create_prefix_node(char *pre_name, int n_seq, rna_seq *head)
{
    pre_node *p_pre_node = (pre_node *) malloc(sizeof(pre_node));
    strcpy(p_pre_node->pre_name, pre_name);
    p_pre_node->n_seq = n_seq;
    p_pre_node->head = head;
    p_pre_node->next = NULL;

    return p_pre_node;
}

void free_prefix_node(pre_node *node)
{
    rna_seq *p;
    
    if (node) {
        p = node->head;
        
        while (p) {
            node->head = p->next;
            free(p);
            p = node->head;
        }

        free(node);
    }

    return;
}

pre_node *find_prefix(pre_list *ptr_list, char *prefix)
{
    pre_node *ptr_node = ptr_list->head;

    while (ptr_node && strcmp(ptr_node->pre_name, prefix))
        ptr_node = ptr_node->next;

    if (!ptr_node)
        return NULL;   // not found
    else
        return ptr_node;
}

void trim_space(char *string)
{
    char *pos;

    //left
    while (*string == ' ')
        string++;

    //right
    pos = string+strlen(string)-1;
    while (*pos == ' ')
        pos--;
    *(pos+1) = '\0';
}

void read_data(char *file_name_u4, char *file_name_u6)
{
    int count = 0;
    FILE *fp_u4, *fp_u6;
    rna_seq *ptr_rna;
    pre_node *ptr_node, *ptr_pos;
    char line[MAXLENGTH], name[256], seq[MAXLENGTH], prefix[16];
    char *pos;

    fp_u4 = fopen(file_name_u4, "r");
    if (!fp_u4) {
        fprintf(stderr, "Failure to open file %s\n", file_name_u4);
        exit(EXIT_FAILURE);
    }
    fp_u6 = fopen(file_name_u6, "r");
    if (!fp_u6) {
        fprintf(stderr, "Failure to open file %s\n", file_name_u6);
        exit(EXIT_FAILURE);
    }

    u4 = (pre_list *) malloc(sizeof(pre_list));
    u4->n_prefix = 0;
    u4->head = NULL;

    u6 = (pre_list *) malloc(sizeof(pre_list));
    u6->n_prefix = 0;
    u6->head = NULL;

    /* read data from U4 stk file */
    while((fgets(line, MAXLENGTH, fp_u4)) != NULL) {
        count++;
        trim_space(line);
        if (line[0]=='#' || line[0]=='\n')     //comment or blank line in stk file
            continue;
        
        pos = strstr(line, " ");
        if (pos == NULL) {
            fprintf(stderr, "%s: line %d    please split name and sequence by space, not by tab!\n",
                    file_name_u4, count);
            exit(EXIT_FAILURE);
        }
        *pos++ = '\0';
        strcpy(name, line);
        
        while (*pos == ' ')
            pos++;
        pos[strlen(pos)-1] = '\0';
        strcpy(seq, pos);
        
        line[3] = '\0';
        strcpy(prefix, line);

        ptr_rna = create_rna_node(name, seq);
        ptr_pos = find_prefix(u4, prefix);
        if (!ptr_pos) {     //@prefix is not in the prefix list
            ptr_node = create_prefix_node(prefix, 0, ptr_rna);
            ptr_node->next = u4->head;
            u4->head = ptr_node;
            ptr_node->head = ptr_rna;

            u4->n_prefix++;
            ptr_node->n_seq++;
        }
        else {
            ptr_rna->next = ptr_pos->head;
            ptr_pos->head = ptr_rna;
            ptr_pos->n_seq++;
        }
    }

    /* read data from U6 stk file */
    count = 0;
    while((fgets(line, MAXLENGTH, fp_u6)) != NULL) {
        count++;
        trim_space(line);
        if (line[0]=='#' || line[0]=='\n')     //comment or structure line in stk file
            continue;
        
        pos = strstr(line, " ");
        if (pos == NULL) {
            fprintf(stderr, "%s: line %d    please split name and sequence by space, not by tab!\n",
                    file_name_u6, count);
            exit(EXIT_FAILURE);
        }
        *pos++ = '\0';
        strcpy(name, line);
        
        while (*pos == ' ')
            pos++;
        pos[strlen(pos)-1] = '\0';
        strcpy(seq, pos);
        
        line[3] = '\0';
        strcpy(prefix, line);

        ptr_rna = create_rna_node(name, seq);
        ptr_pos = find_prefix(u6, prefix);
        if (!ptr_pos) {     //@prefix is not in the prefix list
            ptr_node = create_prefix_node(prefix, 0, ptr_rna);
            ptr_node->next = u6->head;
            u6->head = ptr_node;
            ptr_node->head = ptr_rna;

            u6->n_prefix++;
            ptr_node->n_seq++;
        }
        else {
            ptr_rna->next = ptr_pos->head;
            ptr_pos->head = ptr_rna;
            ptr_pos->n_seq++;
        }
    }
    
    fclose(fp_u6);
    fclose(fp_u4);
}

void free_prefix_list(pre_list *ptr_list)
{
    pre_node *ptr_node;

    if (ptr_list) {
        ptr_node = ptr_list->head;
        
        while (ptr_node) {
            ptr_list->head = ptr_node->next;
            free_prefix_node(ptr_node);
            ptr_node = ptr_list->head;
        }

        free(ptr_list);
    }

    return;
}

void transform_stk2fasta(char *u4_file_name, char *u6_file_name)
{
    FILE *fp_u4, *fp_u6;
    int n4, n6;
    int i, u4_num = 1, u6_num = 1;
    rna_seq *rna_u4, *rna_u6;
    pre_node *cursor, *pre, *ptr_u6;
    char *prefix, *pos, u4_input_name[MAXLENGTH], u6_input_name[MAXLENGTH];

    /* find the intersection of u4 and u6 */
    pre = NULL;
    cursor = u4->head;
    while (cursor) {
        prefix = cursor->pre_name;
        
        if (find_prefix(u6, prefix) == NULL) {  //not found in u6 list
            if (!pre)
                u4->head = cursor->next;
            else
                pre->next = cursor->next;
            
            free_prefix_node(cursor);
            cursor = pre->next;
            u4->n_prefix--;
        }
        else {
            pre = cursor;
            cursor = cursor->next;
        }
    }
    u6->n_prefix = u4->n_prefix;    //note: the number of prefixs in u6, actually may not be equal to @u4->n_prefix

    /* transform stk file to fasta file */
    cursor = u4->head;
    ptr_u6 = u6->head;

    strcpy(u4_input_name, "./data/");
    strcpy(u6_input_name, "./data/");
    strcat(u4_input_name, u4_file_name);
    strcat(u6_input_name, u6_file_name);
    pos = strstr(u4_input_name, ".stk");
    *pos = '\0';
    pos = strstr(u6_input_name, ".stk");
    *pos = '\0';
    strcat(u4_input_name, ".in");
    strcat(u6_input_name, ".in");

    fp_u4 = fopen(u4_input_name, "w");
    fp_u6 = fopen(u6_input_name, "w");
    
    while (cursor) {
        ptr_u6 = find_prefix(u6, cursor->pre_name);
        rna_u4 = cursor->head;

        n4 = cursor->n_seq;
        n6 = ptr_u6->n_seq;

        //write to fasta file format of u4
        while (rna_u4) {
            for (i=0; i<n6; i++) {
                fprintf(fp_u4, ">seq %d: %s\n", u4_num++, rna_u4->name);
                fprintf(fp_u4, "%s\n", rna_u4->seq);
            }

            rna_u4 = rna_u4->next;
        }

        //write to fasta file format of u6
        for (i=0; i<n4; i++) {
            rna_u6 = ptr_u6->head;
            
            while (rna_u6) {
                fprintf(fp_u6, ">seq %d: %s\n", u6_num++, rna_u6->name);
                fprintf(fp_u6, "%s\n", rna_u6->seq);

                rna_u6 = rna_u6->next;
            }
        }

        cursor = cursor->next;
    }
    
    fclose(fp_u6);
    fclose(fp_u4);
    free_prefix_list(u4);
    free_prefix_list(u6);
}
